import os
from functools import partial

import click

from syndicate.exceptions import SyndicateBaseError, AbortedError
from syndicate.commons.log_helper import get_user_logger
from syndicate.core.constants import APPSYNC_TYPE, APPSYNC_DATA_SOURCE_TYPES, \
    APPSYNC_AUTHENTICATION_TYPES, APPSYNC_AUTHORIZATION_TYPES, \
    APPSYNC_RESOLVER_RUNTIMES, APPSYNC_RESOLVER_KINDS, OK_RETURN_CODE, \
    FAILED_RETURN_CODE
from syndicate.core.decorators import return_code_manager
from syndicate.core.generators.appsync import generate_appsync
from syndicate.core.generators.deployment_resources import \
    BaseConfigurationGenerator
from syndicate.core.generators.deployment_resources.appsync_generator import \
    AppSyncDataSourceGenerator, AppSyncAuthorizationGenerator, \
    AppSyncResolverGenerator, AppSyncFunctionGenerator
from syndicate.core.generators.lambda_function import PROJECT_PATH_PARAM
from syndicate.core.helper import resolve_project_path, \
    DictParamType, check_tags, verbose_option, timeit, OptionRequiredIf, \
    ValidRegionParamType, validate_incompatible_options, AliasedCommandsGroup, \
    MultiWordOption, combine_option_classes

OptionCombined = combine_option_classes(OptionRequiredIf, MultiWordOption)

USER_LOG = get_user_logger()


@click.group(name=APPSYNC_TYPE, cls=AliasedCommandsGroup)
@return_code_manager
@click.option('--project-path', '-path', cls=MultiWordOption, nargs=1,
              help="Path to the project root directory. Default value: "
                   "the one from the current config if it exists. "
                   "Otherwise - the current working directory",
              callback=resolve_project_path)
@click.pass_context
def appsync(ctx, project_path):
    """Generates AppSync env and resource meta"""
    if not os.access(project_path, os.F_OK):
        USER_LOG.error(f"The provided path {project_path} doesn't exist")
        return FAILED_RETURN_CODE
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        USER_LOG.error(f"Incorrect permissions for the provided path "
                       f"'{project_path}'")
        return FAILED_RETURN_CODE
    ctx.ensure_object(dict)
    ctx.obj[PROJECT_PATH_PARAM] = project_path
    return OK_RETURN_CODE


@appsync.command(name='api')
@return_code_manager
@click.option('--name', required=True, type=str,
              help="AppSync API name")
@click.option('--project-path', '-path', cls=MultiWordOption, nargs=1,
              help="Path to the project root directory. Default value: "
                   "the one from the current config if it exists. "
                   "Otherwise - the current working directory",
              callback=resolve_project_path)
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@timeit()
def api(name, project_path, tags):
    """
    Generates required environment for AppSync API
    """
    if not os.access(project_path, os.F_OK):
        USER_LOG.error(f"The provided path {project_path} doesn't exist")
        return FAILED_RETURN_CODE
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        USER_LOG.error(f"Incorrect permissions for the provided path "
                       f"'{project_path}'")
        return FAILED_RETURN_CODE
    generate_appsync(name=name,
                     project_path=project_path,
                     tags=tags)
    return OK_RETURN_CODE


@appsync.command(name='data-source')
@return_code_manager
@click.option('--api-name', cls=MultiWordOption,
              required=True, type=str,
              help="AppSync API name to add data source to")
@click.option('--name', required=True, type=str,
              help="Data source name")
@click.option('--description', type=str,
              help="Data source description")
@click.option('--type', type=click.Choice(APPSYNC_DATA_SOURCE_TYPES),
              default='NONE', help="Data source type")
@click.option('--resource-name', cls=OptionCombined, type=str,
              required_if_values=['AWS_LAMBDA', 'AMAZON_DYNAMODB'],
              required_if='type', help="Data source resource name")
@click.option('--region', type=ValidRegionParamType(),
              help="The region where the resource is located. If not "
                   "specified, sets the default value from syndicate config")
@click.option('--service-role-name', cls=OptionCombined, type=str,
              required_if='type',
              required_if_values=['AWS_LAMBDA', 'AMAZON_DYNAMODB'],
              help="The name of the role to access the data source resource")
@verbose_option
@click.pass_context
@timeit()
def data_source(ctx, **kwargs):
    """Adds data source to an existing SyncApp API"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    try:
        generator = AppSyncDataSourceGenerator(**kwargs)
    except Exception as e:
        if isinstance(e, SyndicateBaseError):
            raise click.BadParameter(e)
        raise
    _generate(generator)
    USER_LOG.info(f"Data source '{kwargs['name']}' was added to AppSync API "
                  f"'{kwargs['api_name']}' successfully")
    return OK_RETURN_CODE


@appsync.command(name='function')
@return_code_manager
@click.option('--api-name', cls=MultiWordOption,
              required=True, type=str,
              help="AppSync API name to add function to")
@click.option('--name', required=True, type=str, help="Function name")
@click.option('--description', type=str, help="Function description")
@click.option('--data-source-name', cls=MultiWordOption,
              required=True, type=str,
              help="The name of the data source to associate the function "
                   "with")
@click.option('--runtime', type=click.Choice(APPSYNC_RESOLVER_RUNTIMES),
              required=True, help="Function runtime")
@verbose_option
@click.pass_context
@timeit()
def function(ctx, **kwargs):
    """Adds function to an existing SyncApp API"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    try:
        generator = AppSyncFunctionGenerator(**kwargs)
    except Exception as e:
        if isinstance(e, SyndicateBaseError):
            raise click.BadParameter(e)
        raise
    _generate(generator)
    USER_LOG.info(f"The function '{kwargs['name']}' was added to AppSync API "
                  f"'{kwargs['api_name']}' successfully")
    return OK_RETURN_CODE


@appsync.command(name='resolver')
@return_code_manager
@click.option('--api-name', cls=MultiWordOption,
              required=True, type=str,
              help="AppSync API name to add resolver to")
@click.option('--kind', type=click.Choice(APPSYNC_RESOLVER_KINDS),
              required=True, default='UNIT', is_eager=True,
              help="The kind of resolver.")
@click.option('--type-name', cls=MultiWordOption,
              required=True, type=str,
              help="The name of the type defined in the API schema")
@click.option('--field-name', cls=MultiWordOption,
              required=True, type=str,
              help="The name of the field defined in the API schema to attach "
                   "the resolver to")
@click.option('--data-source-name', cls=OptionCombined, type=str,
              required_if='kind', required_if_values=['UNIT'],
              help="The name of the data source to associate the resolver "
                   "with")
@click.option('--function-name', cls=MultiWordOption,
              type=str, multiple=True,
              callback=partial(
                  validate_incompatible_options,
                  incompatible_options=['data-source-name', 'data_source_name']
              ),
              help="The name of the function to add to the resolver")
@click.option('--runtime', type=click.Choice(APPSYNC_RESOLVER_RUNTIMES),
              required=True, help="Resolver runtime")
@verbose_option
@click.pass_context
@timeit()
def resolver(ctx, **kwargs):
    """Adds resolver to an existing SyncApp API"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    try:
        generator = AppSyncResolverGenerator(**kwargs)
    except Exception as e:
        if isinstance(e, SyndicateBaseError):
            raise click.BadParameter(e)
        raise
    _generate(generator)
    USER_LOG.info(f"The resolver of the type '{kwargs['type_name']}'  for the "
                  f"field '{kwargs['field_name']}' was added to AppSync API "
                  f"'{kwargs['api_name']}' successfully")
    return OK_RETURN_CODE


@appsync.command(name='authorization')
@return_code_manager
@click.option('--api-name', cls=MultiWordOption, required=True,
              type=str, help="AppSync API name to add authorization to")
@click.option('--type', required=True,
              type=click.Choice(APPSYNC_AUTHORIZATION_TYPES),
              help="The authorization type")
@click.option('--auth-type', cls=MultiWordOption, required=True,
              type=click.Choice(APPSYNC_AUTHENTICATION_TYPES),
              help="The authentication type")
@click.option('--resource-name', cls=OptionCombined, type=str,
              required_if_values=['AWS_LAMBDA', 'AMAZON_COGNITO_USER_POOLS'],
              required_if='auth-type',
              help="Authentication provider resource name")
@click.option('--region', type=ValidRegionParamType(),
              help="The region where the authentication provider resource is "
                   "located. If not specified, sets the default value from "
                   "syndicate config")
@verbose_option
@click.pass_context
@timeit()
def authorization(ctx, **kwargs):
    """Adds authorization to an existing SyncApp API"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    try:
        generator = AppSyncAuthorizationGenerator(**kwargs)
    except Exception as e:
        if isinstance(e, SyndicateBaseError):
            raise click.BadParameter(e)
        raise
    _generate(generator)
    USER_LOG.info(f"The '{kwargs['type']}' authorization of type "
                  f"'{kwargs['auth_type']}' was added to AppSync API "
                  f"'{kwargs['api_name']}' successfully")
    return OK_RETURN_CODE


def _generate(generator: BaseConfigurationGenerator):
    """Just some common actions for this module are gathered in here"""
    try:
        generator.write()
    except AbortedError as e:
        raise click.Abort(e)
    except Exception as e:
        if isinstance(e, SyndicateBaseError):
            raise click.BadParameter(e)
        raise Exception(f"An unexpected error occurred: {e}")



import json

import click

from syndicate.exceptions import ResourceNotFoundError, AbortedError, \
    ResourceMetadataError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import API_GATEWAY_TYPE, \
    COGNITO_USER_POOL_AUTHORIZER_TYPE, CUSTOM_AUTHORIZER_KEY, \
    COGNITO_USER_POOL_TYPE, TOKEN_LAMBDA_AUTHORIZER_TYPE, \
    REQUEST_LAMBDA_AUTHORIZER_TYPE
from syndicate.core.generators import (_read_content_from_file,
                                       _write_content_to_file)
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator, BaseConfigurationGenerator


DEFAULT_IDENTITY_SOURCE = 'method.request.header.Authorization'
REQUEST_LAMBDA_AUTHORIZER_DEFAULT_IDENTITY_SOURCE = (
    'method.request.querystring.principal_id, '
    'method.request.querystring.authorization_token, '
    'context.httpMethod, context.resourcePath')
LAMBDAS_ALIAS_KEY = '${lambdas_alias_name}'

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.api_gateway_generator')
USER_LOG = get_user_logger()


class ApiGatewayGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = API_GATEWAY_TYPE
    CONFIGURATION = {
        'deploy_stage': None,
        'dependencies': list,
        'resources': dict,
        'minimum_compression_size': None,
        'tags': dict
    }


class ApiGatewayConfigurationGenerator(BaseConfigurationGenerator):
    """Contains common method for resource and method generators"""

    def __init__(self, **kwargs):
        self.api_gateway_name = kwargs.pop('api_name')
        super().__init__(**kwargs)

    def get_meta_with_resource(self, paths_with_api: list, resource_path: str):
        """Returns the path to deployment resources with api which contains
        resource with given path"""
        for path in paths_with_api:
            deployment_resources = json.loads(_read_content_from_file(path))
            if resource_path in deployment_resources[self.api_gateway_name][
                'resources']:
                _LOG.info(f"Found resource '{resource_path}' in api "
                          f"'{self.api_gateway_name}' in file '{path}'")
                return path, deployment_resources


class ApiGatewayAuthorizerGenerator(ApiGatewayConfigurationGenerator):
    CONFIGURATION = {
        'type': str,
        'identity_source': DEFAULT_IDENTITY_SOURCE,
        'ttl': 300
    }

    def __init__(self, **kwargs):
        self.name = kwargs.pop('name')
        self.type = kwargs.get('type')
        self.provider_name = kwargs.pop('provider_name')
        super().__init__(**kwargs)

        if self.type in [TOKEN_LAMBDA_AUTHORIZER_TYPE,
                         REQUEST_LAMBDA_AUTHORIZER_TYPE]:
            self.CONFIGURATION['lambda_alias'] = LAMBDAS_ALIAS_KEY
            self.CONFIGURATION['lambda_name'] = str
        elif self.type == COGNITO_USER_POOL_AUTHORIZER_TYPE:
            self.CONFIGURATION['user_pools'] = list

    def write(self):
        """Adds an authorizer to the API gateway"""

        paths_with_api = self._get_resource_meta_paths(self.api_gateway_name,
                                                       API_GATEWAY_TYPE)
        if not paths_with_api:
            message = f"Api gateway '{self.api_gateway_name}' was not found"
            _LOG.error(message)
            raise ResourceNotFoundError(message)
        self._resolve_provider()
        USER_LOG.info(f"Adding authorizer '{self.name}' to api "
                      f"'{self.api_gateway_name}'...")
        existing_authorizer_info = self.get_meta_with_authorizer(
            paths_with_api,
            self.name)
        if not existing_authorizer_info:
            path_with_api = paths_with_api[0]
            deployment_resources = json.loads(
                _read_content_from_file(path_with_api))
        else:
            path_with_api, deployment_resources = existing_authorizer_info
            message = f"Authorizer '{self.name}' was found in " \
                      f"api gateway '{self.api_gateway_name}' in file " \
                      f"'{path_with_api}'"
            _LOG.warning(f"Found duplicate while generating meta. "
                         f"{message}")
            if not click.confirm(f"{message} Overwrite?"):
                USER_LOG.warning(f"Skipping authorizer "
                                 f"'{self.name}'")
                raise AbortedError
        USER_LOG.info(f"Adding authorizer '{self.name}' to api "
                      f"'{self.api_gateway_name}'...")
        if not deployment_resources[self.api_gateway_name].get('authorizers'):
            deployment_resources[self.api_gateway_name]['authorizers'] = {}
        deployment_resources[self.api_gateway_name]['authorizers'][
            self.name] = self._resolve_configuration()
        _write_content_to_file(path_with_api,
                               json.dumps(deployment_resources, indent=2))

    def get_meta_with_authorizer(self, paths_with_api: list,
                                 authorizer_name: str):
        """Returns the path to deployment resources with api which contains
        authorizer with given name"""
        for path in paths_with_api:
            deployment_resources = json.loads(_read_content_from_file(path))
            authorizers = deployment_resources[self.api_gateway_name].get(
                'authorizers')
            if not authorizers:
                return
            if authorizer_name in authorizers:
                _LOG.info(f"Found authorizer '{authorizer_name}' in api "
                          f"'{self.api_gateway_name}' in file '{path}'")
                return path, deployment_resources

    def _resolve_provider(self):
        _LOG.info(f"Resolving identity provider '{self.provider_name}'...")
        if self.type in [TOKEN_LAMBDA_AUTHORIZER_TYPE,
                         REQUEST_LAMBDA_AUTHORIZER_TYPE]:
            self._validate_lambda_existence(self.provider_name)
            self._dict['lambda_name'] = self.provider_name
            if self.type == REQUEST_LAMBDA_AUTHORIZER_TYPE:
                self._dict['identity_source'] = \
                    REQUEST_LAMBDA_AUTHORIZER_DEFAULT_IDENTITY_SOURCE
        elif self.type == COGNITO_USER_POOL_AUTHORIZER_TYPE:
            self._validate_cup_existence(self.provider_name)
            self._dict['user_pools'] = [self.provider_name]
        USER_LOG.info(f"Identity provider resolved successfully")

    def _validate_cup_existence(self, cup_name):
        _LOG.info(f"Validating existence of Cognito User Pool: {cup_name}")
        paths_with_cup = self._get_resource_meta_paths(cup_name,
                                                       COGNITO_USER_POOL_TYPE)

        if not paths_with_cup:
            raise ResourceNotFoundError(
                f"Cognito User Pool with name '{cup_name}' not found"
            )
        elif len(paths_with_cup) > 1:
            raise ResourceMetadataError(
                f"More than one Cognito User Pool with the name '{cup_name}' "
                f"found"
            )

        _LOG.info(f"Validation successfully finished, Cognito User Pool "
                  f"exists")


class ApiGatewayResourceGenerator(ApiGatewayConfigurationGenerator):
    CONFIGURATION = {
        'enable_cors': bool
    }

    def __init__(self, **kwargs):
        self.resource_path = kwargs.pop('path')
        super().__init__(**kwargs)

    def write(self):
        """Adds resource to API gateway"""
        paths_with_api = self._get_resource_meta_paths(self.api_gateway_name,
                                                       API_GATEWAY_TYPE)
        if not paths_with_api:
            raise ResourceNotFoundError(
                f"Api gateway '{self.api_gateway_name}' was not found"
            )

        USER_LOG.info(f"Adding resource '{self.resource_path}' to api "
                      f"'{self.api_gateway_name}'...")

        path_with_resources = self.get_meta_with_resource(paths_with_api,
                                                          self.resource_path)
        if not path_with_resources:
            path_with_api = paths_with_api[0]
            deployment_resources = json.loads(
                _read_content_from_file(path_with_api))
        else:
            path_with_api, deployment_resources = path_with_resources
            message = f"Resource '{self.resource_path}' was found in " \
                      f"api gateway '{self.api_gateway_name}' in file " \
                      f"'{path_with_api}'"
            _LOG.warning(f"Found duplicate while generating meta. "
                         f"{message}")
            if not click.confirm(f"{message} Overwrite?"):
                USER_LOG.warning(f"Skipping resource "
                                 f"'{self.resource_path}'")
                raise AbortedError

        deployment_resources[self.api_gateway_name]['resources'][
            self.resource_path] = self._resolve_configuration()
        _write_content_to_file(path_with_api,
                               json.dumps(deployment_resources, indent=2))


class ApiGatewayResourceMethodGenerator(ApiGatewayConfigurationGenerator):
    CONFIGURATION = {
        'authorization_type': "NONE",
        'integration_type': 'mock',
        'lambda_name': None,
        'lamdba_region': None,
        'api_key_required': bool,
        'method_request_parameters': dict,
        'integration_request_body_template': dict,
        'responses': list,
        'integration_responses': list,
        'default_error_pattern': True,
    }

    def __init__(self, **kwargs):
        self.resource_path = kwargs.pop('path')
        self.method = kwargs.pop('method')
        super().__init__(**kwargs)

    def write(self):
        paths_with_api = self._get_resource_meta_paths(self.api_gateway_name,
                                                       API_GATEWAY_TYPE)
        if not paths_with_api:
            raise ResourceNotFoundError(
                f"Api gateway '{self.api_gateway_name}' was not found"
            )

        path_with_resources = self.get_meta_with_resource(paths_with_api,
                                                          self.resource_path)
        if not path_with_resources:
            raise ResourceMetadataError(
                f"Resouce '{self.resource_path}' wasn't found in api "
                f"'{self.api_gateway_name}' definition"
            )

        path_with_api, deployment_resources = path_with_resources

        if self.method in deployment_resources[self.api_gateway_name][
            'resources'][self.resource_path]:
            message = f"Method '{self.method}' was found in resource " \
                      f"'{self.resource_path}' from gateway " \
                      f"'{self.api_gateway_name}' in file '{path_with_api}'"
            _LOG.warning(f"Found duplicate while generating meta. "
                         f"{message}")
            if not click.confirm(f"{message} Overwrite?"):
                USER_LOG.warning(f"Skipping resource "
                                 f"'{self.resource_path}'")
                raise AbortedError

        USER_LOG.info(f"Adding method '{self.method}' to resource "
                      f"'{self.resource_path}' to api "
                      f"'{self.api_gateway_name}'...")

        self._resolve_authorizer(deployment_resources)

        deployment_resources[self.api_gateway_name]['resources'][
            self.resource_path][self.method] = \
            self._resolve_configuration()
        _write_content_to_file(path_with_api,
                               json.dumps(deployment_resources, indent=2))

    def _resolve_configuration(self, defaults_dict=None):
        if self._dict.get('integration_type') == 'lambda':
            lambda_name = self._dict.get('lambda_name')
            self._validate_lambda_existence(lambda_name)
        else:
            if 'lamdba_name' in self._dict:
                self._dict.pop('lambda_name')
        return super()._resolve_configuration()

    def _resolve_authorizer(self, deployment_resources):
        authorization_type = self._dict.get('authorization_type')
        if authorization_type != CUSTOM_AUTHORIZER_KEY:
            return
        _LOG.info(f"Validating existence of authorizer: "
                  f"'{self._dict['authorizer_name']}'")
        authorizers = \
            deployment_resources[self.api_gateway_name].get('authorizers')
        if authorizers:
            if self._dict['authorizer_name'] in authorizers:
                self._dict['authorization_type'] = \
                    self._dict.pop('authorizer_name')
                _LOG.info(f"Validation successfully finished, authorizer "
                          f"'{self._dict['authorization_type']}' exists")
                return

        raise ResourceMetadataError(
            f"Authorizer '{self._dict['authorizer_name']}' wasn't found"
        )



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from syndicate.commons import deep_get
from syndicate.exceptions import ResourceNotFoundError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class DaxResource(BaseResource):
    def __init__(self, dax_conn, iam_conn):
        self.dax_conn = dax_conn
        self.iam_conn = iam_conn

    def create_cluster(self, args):
        return self.create_pool(self._create_cluster_from_meta, args)

    @unpack_kwargs
    def _create_cluster_from_meta(self, name, meta):
        role_name = meta['iam_role_name']
        role_arn = self.iam_conn.check_if_role_exists(role_name)
        if not role_arn:
            message = f'Role {role_name} does not exist; ' \
                      f'Dax cluster {name} failed to be created.'
            _LOG.error(message)
            raise ResourceNotFoundError(message)
        subnet_group_name = meta.get('subnet_group_name')
        subnet_ids = meta.get('subnet_ids') or []
        if subnet_ids:
            _LOG.info(f'Subnet_ids \'{subnet_ids}\' were given. '
                      f'Creating Dax subnet group \'{subnet_group_name}\'')
            self.dax_conn.create_subnet_group(
                subnet_group_name=subnet_group_name,
                subnet_ids=subnet_ids
            )
            _LOG.info(f'Dax subnet group with name {subnet_group_name} '
                      f'was created.')
        elif subnet_group_name:
            _LOG.info(f'Subnet_ids were not given. Assuming that subnet '
                      f'group \'{subnet_group_name}\' already exists')
        response = self.dax_conn.create_cluster(
            cluster_name=name,
            node_type=meta['node_type'],
            replication_factor=meta['replication_factor'],
            iam_role_arn=role_arn,
            subnet_group_name=subnet_group_name,
            cluster_endpoint_encryption_type=meta.get('cluster_endpoint_encryption_type'),
            security_group_ids=meta.get('security_group_ids') or [],
            parameter_group_name=meta.get('parameter_group_name'),
            availability_zones=meta.get('availability_zones') or [],
            tags=meta.get('tags')
        )
        if response:
            _LOG.info(f'Dax cluster \'{name}\' was successfully created')
            return self.describe_cluster(name, meta, response['Cluster'])
        else:
            _LOG.warning(f'Dax cluster \'{name}\' was not created because '
                         f'it already exists')
            return self.describe_cluster(name, meta)

    def describe_cluster(self, name, meta, response=None):
        if not response:
            response = self.dax_conn.describe_cluster(name)
        if response:
            arn = response['ClusterArn']
            del response['ClusterArn']
            return {
                arn: build_description_obj(response, name, meta)
            }
        return {}

    def remove_cluster(self, args):
        return self.create_pool(self._remove_cluster, args)

    @unpack_kwargs
    def _remove_cluster(self, arn, config):
        cluster_name = config['resource_name']
        subnet_group_name = deep_get(config,
                                     ['resource_meta', 'subnet_group_name']) \
            if deep_get(config, ['resource_meta', 'subnet_ids']) else None
        try:
            self.dax_conn.delete_cluster(cluster_name,
                                         log_not_found_error=False)
        except self.dax_conn.client.exceptions.ClusterNotFoundFault:
            _LOG.warning(f'Dax cluster with name \'{cluster_name}\' not found')
            return {arn: config}
        if subnet_group_name:
            self._remove_subnet_group(subnet_group_name)

        return {arn: config}

    def _remove_subnet_group(self, subnet_group_name):
        USER_LOG.info(f"Deleting subnet group '{subnet_group_name}'. "
                      f"Please wait it may take up to 10 minutes.")
        self.dax_conn.delete_subnet_group(subnet_group_name)
        _LOG.info(f"Subnet group '{subnet_group_name}' removed successfully.")



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import EVENT_BRIDGE_SCHEDULE_TYPE


class EventBridgeScheduleGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = EVENT_BRIDGE_SCHEDULE_TYPE
    CONFIGURATION = {
        'schedule_content': {
            'flexible_time_window': dict,
            'schedule_expression': str,
            'target': {
                'arn': str,
                'role_arn': str,
                'input': str,
                'dead_letter_config': dict
            }
        },
        'tags': dict,
    }

    def __init__(self, **kwargs):
        if kwargs.get('mode'):
            kwargs['flexible_time_window'] = {
                'maximum_window_in_minutes': kwargs.pop(
                    'maximum_window_in_minutes', None),
                'mode': kwargs.pop('mode')
            }

        kwargs['target'] = {
            'arn': kwargs.pop('target_arn'),
            'role_arn': kwargs.pop('role_arn')
        }
        if kwargs.get('dead_letter_arn'):
            kwargs['target']['dead_letter_config'] = {
                'arn': kwargs.pop('dead_letter_arn', None)
            }

        resource_name = kwargs.pop('resource_name', {})
        tags = kwargs.pop('tags', {})
        project_path = kwargs.pop('project_path', {})
        schedule_content = kwargs
        kwargs = {'schedule_content': schedule_content, 'tags': tags,
                  'project_path': project_path, 'resource_name': resource_name}
        super().__init__(**kwargs)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from troposphere import apigateway, GetAtt, Ref, Join

from syndicate.exceptions import InvalidValueError, \
    ResourceMetadataError, ParameterError, ResourceNotFoundError
from syndicate.commons.log_helper import get_logger
from syndicate.connection import ApiGatewayConnection
from syndicate.connection.api_gateway_connection import \
    (REQ_VALIDATOR_PARAM_NAME,
     REQ_VALIDATOR_PARAM_VALIDATE_BODY,
     REQ_VALIDATOR_PARAM_VALIDATE_PARAMS,
     RESPONSE_PARAM_ALLOW_HEADERS,
     RESPONSE_PARAM_ALLOW_METHODS,
     RESPONSE_PARAM_ALLOW_ORIGIN)
from syndicate.core.constants import API_GW_DEFAULT_THROTTLING_RATE_LIMIT, \
    API_GW_DEFAULT_THROTTLING_BURST_LIMIT
from syndicate.core.resources.api_gateway_resource import (ApiGatewayResource,
                                                           SUPPORTED_METHODS,
                                                           API_REQUIRED_PARAMS)
from syndicate.core.resources.helper import validate_params
from .cf_lambda_function_converter import CfLambdaFunctionConverter
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import (to_logic_name,
                                  lambda_publish_version_logic_name,
                                  lambda_alias_logic_name,
                                  lambda_function_logic_name,
                                  api_gateway_method_logic_name)

_LOG = get_logger(__name__)


class CfApiGatewayConverter(CfResourceConverter):

    def convert(self, name, meta):
        validate_params(name, meta, API_REQUIRED_PARAMS)
        rest_api = self._convert_rest_api(name=name, meta=meta)
        authorizers_mapping = self._convert_authorizers(meta=meta,
                                                        rest_api=rest_api)
        methods = self._convert_api_resources(
            meta=meta, rest_api=rest_api,
            authorizers_mapping=authorizers_mapping)
        self._convert_deployment(meta=meta, rest_api=rest_api, methods=methods)

    def _convert_api_resources(self, meta, rest_api, authorizers_mapping):
        api_resources = meta.get('resources')
        methods = []
        if api_resources:
            for path, resource_meta in api_resources.items():
                methods.extend(self._process_resource(authorizers_mapping, meta, path, resource_meta, rest_api))
        return methods

    def _process_resource(self, authorizers_mapping, meta, path, resource_meta, rest_api):
        if not path.startswith('/'):
            raise InvalidValueError(
                "API resource must starts with '/', "
                f"but found '{path}'")
        enable_cors = str(resource_meta.get('enable_cors')).lower() == 'true'
        target_resource = self._convert_resource(rest_api=rest_api,
                                                 resource_path=path)
        methods = []
        for method, method_meta in resource_meta.items():
            method_res = apigateway.Method(api_gateway_method_logic_name(path, method))
            if method == 'enable_cors' or method not in SUPPORTED_METHODS:
                continue
            authorization_type = method_meta.get('authorization_type')
            authorizer_id = None
            if authorization_type not in ['NONE', 'AWS_IAM']:
                authorizer_id = authorizers_mapping.get(
                    to_logic_name('ApiGatewayAuthorizer', authorization_type))
                if not authorizer_id:
                    raise ResourceMetadataError(
                        f"Authorizer '{authorization_type}' is not present in "
                        "the build meta."
                    )
                authorization_type = 'CUSTOM'
            self._convert_request_validator(method_meta, method_res, rest_api)

            integration_type = method_meta.get('integration_type')
            integration = self._convert_method_integration(
                integration_type=integration_type,
                method=method,
                method_meta=method_meta,
                path=path,
                rest_api=rest_api)

            method_res.ApiKeyRequired = bool(method_meta.get('api_key_required'))
            method_res.AuthorizationType = authorization_type
            if authorizer_id:
                method_res.AuthorizerId = authorizer_id
            method_res.HttpMethod = method
            if integration:
                method_res.Integration = integration
            if method_meta.get('method_request_models'):
                method_res.RequestModels = method_meta.get('method_request_models')
            if method_meta.get('method_request_parameters'):
                method_res.RequestParameters = method_meta.get('method_request_parameters')
            method_res.ResourceId = Ref(target_resource)
            method_res.RestApiId = Ref(rest_api)
            self.template.add_resource(method_res)

            responses = self._convert_method_responses(enable_cors, meta, method_meta)
            method_res.MethodResponses = responses

            integration_responses = self._convert_integration_responses(enable_cors, meta, method_meta)
            integration.IntegrationResponses = integration_responses
            methods.append(method_res)
        if enable_cors:
            methods.append(self._enable_cors_for_resource(rest_api, path, target_resource))
        return methods

    def _convert_request_validator(self, method_meta, method_resource, rest_api):
        request_validator_meta = method_meta.get('request_validator')
        if request_validator_meta:
            validator_params = \
                ApiGatewayConnection.get_request_validator_params(request_validator_meta)
            request_validator = apigateway.RequestValidator(
                to_logic_name('ApiGatewayRequestValidator', method_resource.title))
            request_validator.Name = validator_params.get(REQ_VALIDATOR_PARAM_NAME)
            request_validator.RestApiId = Ref(rest_api)
            request_validator.ValidateRequestBody = validator_params.get(REQ_VALIDATOR_PARAM_VALIDATE_BODY)
            request_validator.ValidateRequestParameters = validator_params.get(
                REQ_VALIDATOR_PARAM_VALIDATE_PARAMS)
            self.template.add_resource(request_validator)
            method_resource.RequestValidatorId = Ref(request_validator)

    def _convert_rest_api(self, name, meta):
        rest_api = apigateway.RestApi(to_logic_name('ApiGatewayRestApi', name))
        rest_api.Name = name
        binary_media_types = meta.get('binary_media_types')
        if binary_media_types:
            rest_api.BinaryMediaTypes = binary_media_types
        self.template.add_resource(rest_api)
        return rest_api

    def _convert_method_integration(self, integration_type, method, method_meta, path, rest_api):
        integration_builders = {
            'lambda': self._lambda_method_integration,
            'service': self._service_method_integration,
            'mock': self._mock_method_integration,
            'http': self._http_method_integration
        }
        if integration_type not in integration_builders:
            raise InvalidValueError(
                f"'{integration_type}' integration type does not exist."
            )

        integration_builder = integration_builders[integration_type]
        return integration_builder(method, method_meta, path, rest_api)

    def _lambda_method_integration(self, method, method_meta, path, rest_api):
        lambda_name = method_meta.get('lambda_name')
        lambda_arn = self._resolve_lambda_arn(meta=method_meta)
        passthrough_behavior = method_meta.get('integration_passthrough_behavior')
        body_template = method_meta.get('integration_request_body_template')
        enable_proxy = method_meta.get('enable_proxy')
        cache_configuration = method_meta.get('cache_configuration')
        cache_key_parameters = cache_configuration.get(
            'cache_key_parameters') if cache_configuration else None
        integration = apigateway.Integration()
        lambda_uri = self.get_lambda_function_uri(lambda_name=lambda_name)
        if cache_key_parameters:
            integration.CacheKeyParameters = cache_key_parameters
        if method_meta.get('lambda_region'):
            integration.Credentials = method_meta.get('lambda_region')
        integration.IntegrationHttpMethod = 'POST'
        integration.PassthroughBehavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'
        if body_template:
            integration.RequestTemplates = body_template
        integration.Type = 'AWS_PROXY' if enable_proxy else 'AWS'
        integration.Uri = lambda_uri
        api_source_arn = self.get_execute_api_rest_endpoint_arn(
            api_id=Ref(rest_api),
            method=method,
            resource_path=path)
        permission = CfLambdaFunctionConverter.convert_lambda_permission(
            lambda_arn=lambda_arn,
            lambda_name=lambda_name,
            principal='apigateway',
            source_arn=api_source_arn,
            permission_qualifier=rest_api.Name + path + method
        )
        self.template.add_resource(permission)
        return integration

    def _service_method_integration(self, method, method_meta, path, rest_api):
        uri = method_meta.get('uri')
        role = method_meta.get('role')
        passthrough_behavior = method_meta.get('integration_passthrough_behavior')
        body_template = method_meta.get('integration_request_body_template')
        request_parameters = method_meta.get('integration_request_parameters')
        integration_method = method_meta.get('integration_method')
        if not integration_method:
            raise ParameterError(
                f"'integration_method' is not provided for the '{method}' "
                f"'{path}' method of the integration type 'service'"
            )
        credentials = ApiGatewayConnection.get_service_integration_credentials(
            self.config.account_id, role)
        integration = apigateway.Integration()
        integration.Credentials = credentials
        integration.IntegrationHttpMethod = integration_method
        integration.PassthroughBehavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'
        integration.RequestParameters = request_parameters
        integration.RequestTemplates = body_template
        integration.Type = 'AWS'
        integration.Uri = 'arn:aws:apigateway:{0}'.format(uri)
        return integration

    def _mock_method_integration(self, method, method_meta, path, rest_api):
        integration = apigateway.Integration()
        passthrough_behavior = method_meta.get('integration_passthrough_behavior')
        body_template = method_meta.get('integration_request_body_template')
        integration.PassthroughBehavior = passthrough_behavior
        integration.RequestTemplates = body_template
        integration.Type = 'MOCK'
        return integration

    def _http_method_integration(self, method, method_meta, path, rest_api):
        enable_proxy = method_meta.get('enable_proxy')
        integration_method = method_meta.get('integration_method')
        passthrough_behavior = method_meta.get('integration_passthrough_behavior')
        body_template = method_meta.get('integration_request_body_template')
        uri = method_meta.get('uri')
        integration = apigateway.Integration()
        integration.IntegrationHttpMethod = integration_method
        integration.PassthroughBehavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'
        integration.RequestTemplates = body_template
        integration.Type = 'HTTP_PROXY' if enable_proxy else 'HTTP'
        integration.Uri = uri
        return integration

    def _convert_method_responses(self, enable_cors, api_meta, method_meta):
        api_resp = api_meta.get('api_method_responses')
        responses_meta = ApiGatewayResource.init_method_responses(
            api_resp=api_resp,
            method_meta=method_meta)
        responses = []
        if responses_meta:
            for each in responses_meta:
                status_code = each.get('status_code')
                resp_params = each.get('response_parameters', {})
                resp_models = each.get('response_models')
                responses.append(self._convert_method_response(
                    enable_cors=enable_cors,
                    status_code=status_code,
                    response_params=resp_params,
                    response_models=resp_models))
        else:
            responses.append(self._convert_method_response(
                enable_cors=enable_cors))
        return responses

    def _convert_integration_responses(self, enable_cors, api_meta, method_meta):
        api_integration_resp = api_meta.get('api_method_integration_responses')
        integr_responses_meta = ApiGatewayResource.init_integration_method_responses(
            api_integration_resp=api_integration_resp,
            method_meta=method_meta)
        integration_responses = []
        if integr_responses_meta:
            for each in integr_responses_meta:
                status_code = each.get('status_code')
                response_params = each.get('response_parameters')
                response_templates = each.get('response_templates')
                error_regex = each.get('error_regex')
                integration_responses.append(self._convert_integration_method_response(
                    enable_cors=enable_cors,
                    status_code=status_code,
                    response_params=response_params,
                    response_templates=response_templates,
                    selection_pattern=error_regex))
        else:
            integration_responses.append(self._convert_integration_method_response(
                enable_cors=enable_cors))
        return integration_responses

    def _enable_cors_for_resource(self, rest_api, path, target_resource):
        method = 'OPTIONS'
        options_method = apigateway.Method(api_gateway_method_logic_name(path, method))
        options_method.HttpMethod = method
        options_method.AuthorizationType = 'NONE'
        options_method.ResourceId = Ref(target_resource)
        options_method.RestApiId = Ref(rest_api)
        self.template.add_resource(options_method)
        integration = apigateway.Integration()
        integration.RequestTemplates = {
            'application/json': '{"statusCode": 200}'
        }
        integration.Type = 'MOCK'
        options_method.Integration = integration
        method_resp = self._convert_method_response(
            response_params={
                RESPONSE_PARAM_ALLOW_HEADERS: False,
                RESPONSE_PARAM_ALLOW_METHODS: False,
                RESPONSE_PARAM_ALLOW_ORIGIN: False
            })
        options_method.MethodResponses = [method_resp]
        content_types = ("'Content-Type,X-Amz-Date,Authorization,X-Api-Key,"
                         "X-Amz-Security-Token'")
        integr_resp = self._convert_integration_method_response(
            response_params={
                RESPONSE_PARAM_ALLOW_HEADERS: content_types,
                RESPONSE_PARAM_ALLOW_METHODS: "'*'",
                RESPONSE_PARAM_ALLOW_ORIGIN: "'*'"
            })
        integration.IntegrationResponses = [integr_resp]
        return options_method

    def get_lambda_function_uri(self, lambda_name):
        left_part = 'arn:aws:apigateway:{0}:lambda:path' \
                    '/2015-03-31/functions/arn:aws:lambda:{0}:{1}:' \
                    'function:{2}:{3}'.format(
            self.config.region, self.config.account_id, lambda_name,
            self.config.lambdas_alias_name)
        right_part = '/invocations'
        return Join('', [left_part, right_part])

    def get_execute_api_rest_endpoint_arn(self, api_id, method, resource_path,
                                          stage='*'):
        left_part = 'arn:aws:execute-api:{0}:{1}:'.format(
            self.config.region, self.config.account_id)
        right_part = '/{0}/{1}{2}'.format(stage, method, resource_path)
        return Join('', [left_part, api_id, right_part])

    @staticmethod
    def _convert_method_response(enable_cors=False, status_code=None,
                                 response_params=None, response_models=None):
        response_allow_origin = "response.header.Access-Control-Allow-Origin"
        method_allow_origin = "method.{0}".format(response_allow_origin)
        if not response_params:
            response_params = {}
        if enable_cors:
            response_params[method_allow_origin] = False
        if not response_models:
            response_models = {'application/json': 'Empty'}
        return apigateway.MethodResponse(
            ResponseModels=response_models,
            ResponseParameters=response_params,
            StatusCode=status_code if status_code else '200')

    @staticmethod
    def _convert_integration_method_response(enable_cors=False,
                                             status_code=None,
                                             response_params=None,
                                             response_templates=None,
                                             selection_pattern=None):
        response_allow_origin = "response.header.Access-Control-Allow-Origin"
        method_allow_origin = "method.{0}".format(response_allow_origin)
        if not response_params:
            response_params = {}
        if enable_cors:
            response_params[method_allow_origin] = "'*'"
        if not response_templates:
            response_templates = {'application/json': ''}
        response = apigateway.IntegrationResponse(
            ResponseParameters=response_params,
            ResponseTemplates=response_templates,
            StatusCode=status_code if status_code else '200')
        if selection_pattern:
            response.SelectionPattern = selection_pattern
        return response

    def _convert_authorizers(self, meta, rest_api):
        authorizers_mapping = {}
        authorizers_meta = meta.get('authorizers', {})
        if authorizers_meta:
            authorizers = list()
            for key, val in authorizers_meta.items():
                authorizer_name = to_logic_name('ApiGatewayAuthorizer', key)
                authorizer = apigateway.Authorizer(authorizer_name)
                authorizer.Name = authorizer_name
                authorizer.RestApiId = Ref(rest_api)
                authorizer.Type = val.get('type')
                authorizer.AuthorizerResultTtlInSeconds = val.get('ttl')
                authorizer.IdentitySource = val.get('identity_source')

                lambda_arn = self._resolve_lambda_arn(meta=val)
                lambda_name = val.get('lambda_name')

                authorizer_uri = self.get_lambda_function_uri(lambda_name=lambda_name)
                authorizer.AuthorizerUri = authorizer_uri
                self.template.add_resource(authorizer)
                authorizers.append(authorizer)

                permission = CfLambdaFunctionConverter.convert_lambda_permission(
                    lambda_arn=lambda_arn,
                    lambda_name=lambda_name,
                    principal='apigateway'
                )
                self.template.add_resource(permission)
            authorizers_mapping = {x.Name: Ref(x) for x in authorizers}
        return authorizers_mapping

    def _convert_deployment(self, meta, rest_api, methods):
        stage_name = \
            ApiGatewayResource.get_deploy_stage_name(meta.get('deploy_stage'))
        cache_cluster_configuration = meta.get('cluster_cache_configuration')
        cache_cluster_enabled = cache_cluster_configuration.get(
            'cache_enabled') if cache_cluster_configuration else None
        cache_size_value = cache_cluster_configuration.get(
            'cache_size') if cache_cluster_configuration else None
        cache_cluster_size = \
            str(cache_size_value) if cache_size_value else None
        throttling_cluster_configuration = meta.get(
            'cluster_throttling_configuration')
        throttling_cluster_enabled = throttling_cluster_configuration.get(
            'throttling_enabled') if throttling_cluster_configuration else None
        deployment = apigateway.Deployment(
            to_logic_name('ApiGatewayDeployment', rest_api.title))
        stage = apigateway.StageDescription()
        if str(cache_cluster_enabled).lower() == 'true':
            stage.CacheClusterEnabled = True
        if cache_cluster_size:
            stage.CacheClusterSize = cache_cluster_size
        if cache_cluster_enabled:
            cache_ttl_sec = cache_cluster_configuration.get('cache_ttl_sec')
            if cache_ttl_sec is not None:
                stage.CacheTtlInSeconds = cache_ttl_sec
        if throttling_cluster_enabled:
            stage.ThrottlingRateLimit = throttling_cluster_configuration.get(
                'throttling_rate_limit', API_GW_DEFAULT_THROTTLING_RATE_LIMIT)
            stage.ThrottlingBurstLimit = throttling_cluster_configuration.get(
                'throttling_burst_limit', API_GW_DEFAULT_THROTTLING_BURST_LIMIT)
        deployment.StageDescription = stage
        deployment.RestApiId = Ref(rest_api)
        deployment.StageName = stage_name
        deployment.DependsOn = methods
        self.template.add_resource(deployment)

    def _convert_resource(self, rest_api, resource_path):
        parent_resource_id = GetAtt(rest_api.title, 'RootResourceId')
        resource_path = resource_path[1:]
        resource_path_split = resource_path.split('/')
        target_resource = None
        for path_part in resource_path_split:
            api_resource_name = to_logic_name(
                'ApiGatewayResource',
                ''.join(resource_path.partition(path_part)[:2]))
            existing_resource = self.get_resource(api_resource_name)
            if existing_resource:
                target_resource = existing_resource
                parent_resource_id = Ref(existing_resource)
                continue
            resource = apigateway.Resource(api_resource_name)
            resource.ParentId = parent_resource_id
            resource.PathPart = path_part
            resource.RestApiId = Ref(rest_api)
            self.template.add_resource(resource)
            target_resource = resource
            parent_resource_id = Ref(resource)
        return target_resource

    def _resolve_lambda_arn(self, meta):
        lambda_name = meta.get('lambda_name')
        lambda_version = meta.get('lambda_version')
        lambda_alias = meta.get('lambda_alias')
        lambda_arn = None
        if lambda_alias:
            alias_resource = self.get_resource(
                lambda_alias_logic_name(function_name=lambda_name,
                                        alias=lambda_alias))
            if alias_resource:
                lambda_arn = Ref(alias_resource)
        elif lambda_version:
            version_resource = self.get_resource(
                lambda_publish_version_logic_name(
                    function_name=lambda_name))
            if version_resource:
                lambda_arn = Ref(version_resource)
        else:
            function_resource = self.get_resource(
                lambda_function_logic_name(function_name=lambda_name))
            if function_resource:
                lambda_arn = GetAtt(function_resource.title, 'Arn')
        if lambda_arn is None:
            function_ref = lambda_name
            if lambda_alias:
                function_ref += ':' + lambda_alias
            elif lambda_version:
                function_ref += ':' + lambda_version
            raise ResourceNotFoundError(
                f"Lambda function '{function_ref}' is not present in build "
                f"meta."
            )
            # Or lookup existing lambda function in AWS?
            # lambda_service = self.resources_provider.lambda_resource()
            # function_arn = \
            #     lambda_service.resolve_lambda_arn_by_version_and_alias(
            #         name=lambda_name,
            #         alias=lambda_alias,
            #         version=lambda_version)
        return lambda_arn



import json
import os
from functools import partial

import click

from syndicate.core.generators.deployment_resources.rds_generator import \
    RDSDBClusterGenerator, RDSDBInstanceGenerator
from syndicate.exceptions import AbortedError,  SyndicateBaseError
from syndicate.commons.log_helper import get_user_logger
from syndicate.core.constants import (
    S3_BUCKET_ACL_LIST, API_GW_AUTHORIZER_TYPES, CUSTOM_AUTHORIZER_KEY,
    EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS, EC2_LT_RESOURCE_TAGS,
    FAILED_RETURN_CODE, OK_RETURN_CODE,
)
from syndicate.core.decorators import return_code_manager
from syndicate.core.generators.deployment_resources import *
from syndicate.core.generators.deployment_resources.api_gateway_generator import \
    ApiGatewayAuthorizerGenerator

from syndicate.core.generators.deployment_resources.ec2_launch_template_generator import \
    EC2LaunchTemplateGenerator
from syndicate.core.generators.deployment_resources.eventbridge_schedule import \
    EventBridgeScheduleGenerator
from syndicate.core.generators.deployment_resources.firehose_generator import \
    FirehoseGenerator
from syndicate.core.generators.lambda_function import PROJECT_PATH_PARAM
from syndicate.core.helper import (
    OptionRequiredIf, check_tags,
    validate_authorizer_name_option, verbose_option, validate_api_gw_path,
    DictParamType, DeepDictParamType, validate_incompatible_options,
    AliasedCommandsGroup, MultiWordOption, combine_option_classes,
)
from syndicate.core.helper import ValidRegionParamType
from syndicate.core.helper import validate_bucket_name
from syndicate.core.helper import resolve_project_path, timeit

GENERATE_META_GROUP_NAME = 'meta'
dynamodb_type_param = click.Choice(['S', 'N', 'B'])

RDS_INSTANCE_DB_CLUSTER_INCOMPATIBLE_OPTIONS = [
    'engine', 'engine-version', 'master-username', 'master-password',
    'database-name', 'port', 'vpc-security-group-ids', 'availability-zone']

OptionCombined = combine_option_classes(OptionRequiredIf, MultiWordOption)

USER_LOG = get_user_logger()


@click.group(name=GENERATE_META_GROUP_NAME, cls=AliasedCommandsGroup)
@return_code_manager
@click.option('--project-path', '-path',
              cls=MultiWordOption, nargs=1,
              help="Path to the project root directory. Default value: "
                   "the one from the current config if it exists. "
                   "Otherwise - the current working directory",
              callback=resolve_project_path)
@click.pass_context
def meta(ctx, project_path):
    """Generates deployment resources templates"""
    if not os.access(project_path, os.F_OK):
        USER_LOG.error(f"The provided path {project_path} doesn't exist")
        return FAILED_RETURN_CODE
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        USER_LOG.error(f"Incorrect permissions for the provided path "
                       f"'{project_path}'")
        return FAILED_RETURN_CODE
    ctx.ensure_object(dict)
    ctx.obj[PROJECT_PATH_PARAM] = project_path
    return OK_RETURN_CODE


@meta.command(name='dax-cluster')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption,
              required=True, type=str, help="Dax cluster name")
@click.option('--node-type',
              cls=MultiWordOption,
              required=True, type=str,
              help="The node type for the nodes in the cluster")
@click.option('--iam-role-name',
              cls=MultiWordOption,
              required=True, type=str,
              help="Role name to access DynamoDB tables")
@click.option('--subnet-group-name',
              cls=MultiWordOption,
              required=True, type=str,
              help='The name of the subnet group to be used for the '
                   'replication group')
@click.option('--subnet-ids',
              cls=MultiWordOption, type=str, multiple=True,
              help='Subnet ids to create a subnet group from. Don\'t specify '
                   'in case of using existing subnet group')
@click.option('--cluster-endpoint-encryption-type',
              cls=MultiWordOption,
              type=click.Choice(['NONE', 'TLS']), default='TLS',
              help='The encryption type of the cluster\'s endpoint. '
                   'The default value is \'TLS\'')
@click.option('--parameter-group-name',
              cls=MultiWordOption, type=str,
              help='The parameter group to be associated with the DAX cluster')
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def dax_cluster(ctx, **kwargs):
    """Generated dax cluster deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DaxClusterGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f'Dax cluster \'{kwargs["resource_name"]}\' was '
                  f'successfully generated')
    return OK_RETURN_CODE


@meta.command(name='dynamodb')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, required=True, type=str,
              help="DynamoDB table name")
@click.option('--hash-key-name',
              cls=MultiWordOption, required=True, type=str,
              help="DynamoDB table hash key")
@click.option('--hash-key-type',
              cls=MultiWordOption, required=True,
              type=click.Choice(['S', 'N', 'B']),
              help="DynamoDB hash key type")
@click.option('--sort-key-name',
              cls=MultiWordOption, type=str,
              help="DynamoDB sort key. If not specified, the table will have "
                   "only a hash key")
@click.option('--sort-key-type',
              type=dynamodb_type_param,
              cls=OptionCombined, required_if='sort-key-name',
              help="Required if sort key name is specified")
@click.option('--billing-mode',
              cls=MultiWordOption, required=True,
              type=click.Choice(['PROVISIONED', 'PAY_PER_REQUEST']),
              default='PAY_PER_REQUEST',
              help="Controls how you are charged for read and write "
                   "throughput")
@click.option('--read-capacity',
              cls=MultiWordOption, type=int,
              help="The maximum number of strongly consistent reads that can"
                   "be performed per second in PROVISIONED billing mode. "
                   "Maximum number of read request units in PAY_PER_REQUEST "
                   "billing mode. If not specified, sets the default value "
                   "to 1")
@click.option('--write-capacity',
              cls=MultiWordOption, type=int,
              help="The maximum number of writing processes consumed per"
                   "second in PROVISIONED billing mode. "
                   "Maximum number of write request units in PAY_PER_REQUEST "
                   "billing mode. If not specified, sets the default value "
                   "to 1")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def dynamodb(ctx, **kwargs):
    """Generates dynamoDB deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DynamoDBGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Table '{kwargs['resource_name']}' was added successfully!")
    return OK_RETURN_CODE


@meta.command(name='dynamodb-global-index')
@return_code_manager
@click.option('--table-name',
              cls=MultiWordOption, required=True, type=str,
              help="DynamoDB table name to add index to")
@click.option('--name', required=True, type=str,
              help="Index name")
@click.option('--index-key-name',
              cls=MultiWordOption, required=True, type=str,
              help="Index hash key")
@click.option('--index-key-type',
              cls=MultiWordOption, required=True,
              type=dynamodb_type_param,
              help='Hash key index type')
@click.option('--index-sort-key_name',
              cls=MultiWordOption, type=str,
              help='Index sort key')
@click.option('--index-sort-key-type',
              cls=OptionCombined, type=dynamodb_type_param,
              required_if='index-sort-key-name',
              help="Sort key type")
@verbose_option
@click.pass_context
@timeit()
def dynamodb_global_index(ctx, **kwargs):
    """Adds dynamodb global index to existing dynamodb table"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DynamoDBGlobalIndexGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Global index '{kwargs['name']}' was added successfully")
    return OK_RETURN_CODE


@meta.command(name='dynamodb-autoscaling')
@return_code_manager
@click.option('--table-name',
              cls=MultiWordOption, type=str, required=True,
              help="DynamoDB table name to add autoscaling to")
@click.option('--policy-name',
              cls=MultiWordOption, type=str, required=True,
              help="Autoscaling policy name")
@click.option('--min-capacity',
              cls=MultiWordOption, type=click.IntRange(min=1),
              help="Minimum capacity level. If not specified, sets the default"
                   " value to 1")
@click.option('--max-capacity',
              cls=MultiWordOption, type=click.IntRange(min=1),
              help="Maximum capacity level. If not specified, sets the default"
                   " value to 10")
@click.option('--target-utilization',
              cls=MultiWordOption,
              type=click.IntRange(min=20, max=90),
              help="Target utilization in autoscaling. If not specified, sets "
                   "the default value to 70 %")
@click.option('--scale-in-cooldown',
              cls=MultiWordOption, type=click.IntRange(min=0),
              help="Scaling policy value of in cooldown in seconds. Is not "
                   "specified, sets the default value to 60")
@click.option('--scale-out-cooldown',
              cls=MultiWordOption, type=click.IntRange(min=0),
              help="Scaling policy value of out cooldown in seconds. Is not "
                   "specified, sets the default value to 60")
@click.option('--dimension', type=str,
              help="Autoscaling dimension. If not specified, sets the default"
                   "value to 'dynamodb:table:ReadCapacityUnits'")
@click.option('--role-name',
              cls=MultiWordOption, type=str,
              help="The name of the role, which performs autoscaling. If not "
                   "specified, sets the value to default service linked role: "
                   "'AWSServiceRoleForApplicationAutoScaling_DynamoDBTable'")
@verbose_option
@click.pass_context
@timeit()
def dynamodb_autoscaling(ctx, **kwargs):
    """Adds autoscaling settings to existing dynamodb table"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DynamoDBAutoscalingGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Autoscaling setting to table '{kwargs['table_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='s3-bucket')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, required=True, type=str,
              help="S3 bucket name", callback=validate_bucket_name)
@click.option('--location', type=ValidRegionParamType(),
              help="The region where the bucket is created, the default value "
                   "is the region set in syndicate config")
@click.option('--acl', type=click.Choice(S3_BUCKET_ACL_LIST),
              help="The channel ACL to be applied to the bucket. If not "
                   "specified, sets the default value to 'private'")
@click.option('--block-public-acls',
              cls=MultiWordOption,
              type=bool, required=False, is_eager=True,
              help='Specifies whether Amazon S3 should block public access '
                   'control lists (ACLs) for this bucket and objects in this '
                   'bucket. Default value is True')
@click.option('--ignore-public-acls',
              cls=MultiWordOption,
              type=bool, required=False, is_eager=True,
              help='Specifies whether Amazon S3 should ignore public ACLs for '
                   'this bucket and objects in this bucket. Default value '
                   'is True')
@click.option('--block-public-policy',
              cls=MultiWordOption,
              type=bool, required=False, is_eager=True,
              help='Specifies whether Amazon S3 should block public bucket '
                   'policies for this bucket. Default value is True')
@click.option('--restrict-public-buckets',
              cls=MultiWordOption,
              type=bool, required=False, is_eager=True,
              help='Specifies whether Amazon S3 should restrict public bucket '
                   'policies for this bucket. Default value is True')
@click.option('--static-website-hosting',
              cls=MultiWordOption, type=bool, required=False,
              help='Specifies whether the S3 bucket should be configured for '
                   'static WEB site hosting. Default value is False')
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def s3_bucket(ctx, **kwargs):
    """Generates s3 bucket deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = S3Generator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"S3 bucket '{kwargs['resource_name']}' was "
                  f"added successfully!")
    return OK_RETURN_CODE


@meta.command(name='api-gateway')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, required=True, type=str,
              help="Api gateway name")
@click.option('--deploy-stage',
              cls=MultiWordOption, required=True, type=str,
              help="The stage to deploy the API")
@click.option('--minimum-compression-size',
              cls=MultiWordOption,
              type=click.IntRange(min=0, max=10 * 1024 * 1024),
              help="Compression size for api gateway. If not specified, "
                   "compression will be disabled")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def api_gateway(ctx, **kwargs):
    """Generates api gateway deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Api gateway '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='web-socket-api-gateway')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, required=True, type=str,
              help="Api gateway name")
@click.option('--deploy-stage',
              cls=MultiWordOption, required=True, type=str,
              help="The stage to deploy the API")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def web_socket_api_gateway(ctx, **kwargs):
    """Generates web socket api gateway deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = WebSocketApiGatewayGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Api gateway '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='api-gateway-authorizer')
@return_code_manager
@click.option('--api-name',
              cls=MultiWordOption, required=True, type=str,
              help="Api gateway name to add authorizer to")
@click.option('--name', required=True, type=str,
              help="Authorizer name")
@click.option('--type', type=click.Choice(API_GW_AUTHORIZER_TYPES),
              required=True, help="Authorizer type. 'TOKEN' for a Lambda "
                                  "function using a single authorization "
                                  "token submitted in a custom header, "
                                  "'REQUEST' for a Lambda function using "
                                  "incoming request parameters, and "
                                  "'COGNITO_USER_POOLS' for using an Amazon "
                                  "Cognito user pool")
@click.option('--provider-name',
              cls=MultiWordOption, type=str, required=True,
              help="Identity provider name")
@verbose_option
@click.pass_context
@timeit()
def api_gateway_authorizer(ctx, **kwargs):
    """Adds authorizer to an existing api gateway"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayAuthorizerGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Authorizer '{kwargs['name']}' was added to API gateway "
                  f"'{kwargs['api_name']}' successfully")
    return OK_RETURN_CODE


@meta.command(name='api-gateway-resource')
@return_code_manager
@click.option('--api-name',
              cls=MultiWordOption, required=True, type=str,
              help="Api gateway name to add resource to")
@click.option('--path', required=True, callback=validate_api_gw_path,
              help="Resource path to create")
@click.option('--enable-cors',
              cls=MultiWordOption, type=bool,
              help="Enables CORS on the resource method. If not specified, "
                   "sets the default value to False")
@verbose_option
@click.pass_context
@timeit()
def api_gateway_resource(ctx, **kwargs):
    """Adds resource to existing api gateway"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayResourceGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Resource '{kwargs['path']}' was added to API gateway "
                  f"'{kwargs['api_name']}' successfully")
    return OK_RETURN_CODE


@meta.command(name='api-gateway-resource-method')
@return_code_manager
@click.option('--api-name',
              cls=MultiWordOption, required=True, type=str,
              help="Api gateway name to add method to")
@click.option('--path', required=True, callback=validate_api_gw_path,
              help="Resource path to add method to")
@click.option('--method', required=True,
              type=click.Choice(['POST', 'GET', 'DELETE', 'PUT', 'HEAD',
                                 'PATCH', 'ANY']),
              help="Resource method to add")
@click.option('--integration-type',
              cls=MultiWordOption, type=str,
              help="The resource which the method is connected to: "
                   "[lambda|service|http|mock]. If not specified, sets the "
                   "default value to 'mock'")
@click.option('--lambda-name',
              cls=MultiWordOption, type=str,
              help="Lambda name. Required if integration type is lambda")
@click.option('--lambda-region',
              cls=MultiWordOption, type=ValidRegionParamType(),
              help="The region where the lambda is located. If not specified, "
                   "sets the default value from syndicate config")
@click.option('--authorization-type',
              cls=MultiWordOption, is_eager=True,
              type=click.Choice(["NONE", "AWS_IAM", CUSTOM_AUTHORIZER_KEY]),
              help="The method's authorization type. If not specified, sets "
                   "the default value to 'NONE'")
@click.option('--authorizer-name',
              cls=MultiWordOption, type=str,
              callback=validate_authorizer_name_option,
              help="The method's authorizer name can be used only with "
                   "'--authorization_type' 'CUSTOM'")
@click.option('--api-key-required',
              cls=MultiWordOption, type=bool,
              help="Specifies whether the method requires a valid API key. "
                   "If not specified, the default value is set to False")
@verbose_option
@click.pass_context
@timeit()
def api_gateway_resource_method(ctx, **kwargs):
    """Adds a method to existing api gateway resource"""

    if kwargs.get('integration_type') == 'lambda' \
            and not kwargs.get('lambda_name'):
        raise click.MissingParameter(
            "Lambda name is required if the integration type is 'lambda'",
            param_type='option', param_hint='lambda_name')

    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayResourceMethodGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Method '{kwargs['method']}' was added to API gateway "
                  f"resource '{kwargs['path']}' successfully")
    return OK_RETURN_CODE


@meta.command(name='iam-policy')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, required=True, type=str,
              help='IAM policy name')
@click.option('--policy-content',
              cls=MultiWordOption,
              help='The path to JSON file with IAM policy content. '
                   'If not specified, template value will be set',
              type=click.File(mode='r'))
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def iam_policy(ctx, **kwargs):
    """Generates IAM policy deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    if kwargs['policy_content']:
        try:
            kwargs['policy_content'] = json.load(kwargs['policy_content'])
        except json.decoder.JSONDecodeError as e:
            raise click.BadParameter(str(e), param_hint='policy_content')
    generator = IAMPolicyGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Iam policy '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='iam-role')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, required=True, type=str,
              help="IAM role name")
@click.option('--principal-service',
              cls=MultiWordOption, required=True, type=str,
              help="The service which will use the role")
@click.option('--predefined-policies',
              cls=MultiWordOption, type=str, multiple=True,
              help="Managed IAM policies list")
@click.option('--custom-policies',
              cls=MultiWordOption, type=str, multiple=True,
              help="Customer AWS policies names")
@click.option('--allowed-accounts',
              cls=MultiWordOption, type=str, multiple=True,
              help="The list of accounts, which can assume the role")
@click.option('--external-id',
              cls=MultiWordOption,
              type=str, help="External ID in role")
@click.option('--instance-profile',
              cls=MultiWordOption, type=bool,
              help="If true, instance profile with role name is created")
@click.option('--permissions-boundary',
              cls=MultiWordOption, type=str,
              help="The name or the ARN of permissions boundary policy to "
                   "attach to this role")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def iam_role(ctx, **kwargs):
    """Generates IAM role deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = IAMRoleGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Iam role '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='kinesis-stream')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Kinesis stream name")
@click.option('--shard-count',
              cls=MultiWordOption, type=int, required=True,
              help="Number of shards that the stream uses")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def kinesis_stream(ctx, **kwargs):
    """Generates kinesis stream deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = KinesisStreamGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Kinesis stream '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='sns-topic')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="SNS topic name")
@click.option('--region', type=ValidRegionParamType(allowed_all=True),
              required=True, help="Where the topic should be deployed")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def sns_topic(ctx, **kwargs):
    """Generates sns topic deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = SNSTopicGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"SNS topic '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='step-function')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Step function name")
@click.option('--iam-role',
              cls=MultiWordOption, type=str, required=True,
              help="IAM role to use for this state machine")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def step_function(ctx, **kwargs):
    """Generate step function deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = StepFunctionGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Step function '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='step-function-activity')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Step function activity name")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def step_function_activity(ctx, **kwargs):
    """Generates step function activity deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = StepFunctionActivityGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Step function activity '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='ec2-instance')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Instance name")
@click.option('--key-name',
              cls=MultiWordOption, type=str, required=True,
              help="SSH key to access the instance")
@click.option('--image-id',
              cls=MultiWordOption, type=str, required=True,
              help="Image id to create the instance from")
@click.option('--instance-type',
              cls=MultiWordOption, type=str,
              help="Instance type. Default type: t2.micro")
@click.option('--disable-api-termination',
              cls=MultiWordOption, type=bool,
              help="Api termination protection. Default value is True")
@click.option('--security-group-ids',
              cls=MultiWordOption, type=str, multiple=True,
              help="Security group ids")
@click.option('--security-group-names',
              cls=MultiWordOption, type=str, multiple=True,
              help="Security group names")
@click.option('--availability-zone',
              cls=MultiWordOption, type=str,
              help="Instance availability zone")
@click.option('--subnet-id',
              type=str, cls=OptionCombined,
              required_if="availability-zone",
              help="Subnet ID (required if availability zone is set)")
@click.option('--userdata-file',
              cls=MultiWordOption, type=str,
              help="File path to userdata (file relative pathname from the"
                   "directory which is set up in the env variable 'SDCT_CONF'")
@click.option('--iam-role',
              cls=MultiWordOption, type=str,
              help="Instance IAM role")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def ec2_instance(ctx, **kwargs):
    """Generates ec2 instance deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = EC2InstanceGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"EC2 instance '{kwargs['resource_name']}' was added"
                  f"successfully")
    return OK_RETURN_CODE


@meta.command(name='ec2-launch-template')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Launch template name")
@click.option('--image-id',
              cls=MultiWordOption, type=str, required=True,
              help="The ID of the AMI")
@click.option('--key-name',
              cls=MultiWordOption, type=str,
              help="The name of the key pair")
@click.option('--instance-type',
              cls=MultiWordOption, type=str,
              help="Instance type")
@click.option('--security-group-ids',
              cls=MultiWordOption, type=str, multiple=True,
              help="Security group ids")
@click.option('--security-group-names', type=str, multiple=True,
              help="Security group names")
@click.option('--userdata-file',
              cls=MultiWordOption, type=str,
              help="File path to userdata (can be specified as a relative "
                   "path to the project path)")
@click.option('--iam-role',
              cls=MultiWordOption, type=str,
              help="Instance IAM role")
@click.option('--imds-version',
              cls=MultiWordOption,
              type=click.Choice(EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS),
              help="IMDS version")
@click.option('--version-description',
              cls=MultiWordOption, type=str,
              help="A description for the version of the launch template")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The ec2 launch template tags')
@click.option('--resource-tags',
              cls=MultiWordOption,
              type=DeepDictParamType(), multiple=True,
              help=f'The resource tags. You can specify tags for the '
                   f'following {EC2_LT_RESOURCE_TAGS}. To tag a resource '
                   f'after it has been created')
@verbose_option
@click.pass_context
@timeit()
def ec2_launch_template(ctx, **kwargs):
    """Generates ec2_launch_template deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]

    if kwargs["resource_tags"]:
        kwargs["resource_tags"] = \
            {k: v for d in kwargs["resource_tags"] for k, v in d.items()}
        valid_keys = set(EC2_LT_RESOURCE_TAGS)
        provided_keys = set(kwargs['resource_tags'].keys())
        # Check if provided keys are all valid
        if not provided_keys <= valid_keys:
            invalid_keys = provided_keys - valid_keys
            raise click.BadParameter(
                f"Invalid resource tag keys provided: '{invalid_keys}'. "
                f"Allowed keys are: '{valid_keys}'"
            )

    generator = EC2LaunchTemplateGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"ec2_launch_template '{kwargs['resource_name']}' was added "
                  f"successfully")
    return OK_RETURN_CODE


@meta.command(name='sqs-queue')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="SQS queue name")
@click.option('--region', type=ValidRegionParamType(),
              help="The region where the queue is deployed. Default value is "
                   "the one from syndicate config")
@click.option('--fifo-queue',
              cls=MultiWordOption, type=bool,
              help="If True, the queue is FIFO. Default value is False")
@click.option('--visibility-timeout',
              cls=MultiWordOption,
              type=click.IntRange(min=0, max=43200),
              help="The visibility timeout for the queue. Default value is 30")
@click.option('--delay-seconds',
              cls=MultiWordOption,
              type=click.IntRange(min=0, max=900),
              help="The length of time in seconds for which the delivery "
                   "of all the messages in the queue is delayed. Default "
                   "value is 0")
@click.option('--maximum-message-size',
              cls=MultiWordOption,
              type=click.IntRange(min=1024, max=262144),
              help="The limit of how many bytes a message can contain before "
                   "Amazon SQS rejects it. Default value is 1024")
@click.option('--message-retention-period',
              cls=MultiWordOption,
              type=click.IntRange(min=60, max=1209600),
              help="The length of time in seconds for which Amazon SQS "
                   "retains a message. Default value is 60")
@click.option('--receive-message-wait-time-seconds',
              cls=MultiWordOption,
              type=click.IntRange(min=0, max=20),
              help="The length of time in seconds for which a 'ReceiveMessage'"
                   " action waits for a message to arrive")
@click.option('--dead-letter-target-arn',
              cls=MultiWordOption, type=str,
              help="Arn of a dead-letter queue Amazon SQS moves messages "
                   "after the value of maxReceiveCount is exceeded")
@click.option('--max-receive-count', type=click.IntRange(min=1, max=1000),
              help="The number of times a message is delivered to the source "
                   "queue before being moved to the dead-letter queue. "
                   "Required if 'dead-letter-target-arn' is specified",
              cls=OptionCombined, required_if='dead-letter-target-arn')
@click.option('--kms-master-key-id', type=str,
              help="The id of an AWS-managed customer master key (CMK) for "
                   "Amazon SQS or a custom CMK")
@click.option('--kms-data-key-reuse-period-seconds',
              cls=MultiWordOption,
              type=click.IntRange(min=60, max=86400),
              help="The length of time in seconds for which Amazon SQS can "
                   "reuse a data key to encrypt or decrypt messages before "
                   "calling AWS KMS again")
@click.option('--content-based-deduplication',
              cls=MultiWordOption, type=bool,
              help="Enables content-based deduplication")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def sqs_queue(ctx, **kwargs):
    """Generates sqs queue deployment deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = SQSQueueGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"SQS queue '{kwargs['resource_name']}' was added "
                  f"successfully")
    return OK_RETURN_CODE


@meta.command(name="sns-application")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="The name of the sns application")
@click.option('--platform', required=True,
              type=click.Choice(['GCM', 'ADM', 'APNS', 'APNS_SANDBOX']),
              help="SNS application platform")
@click.option('--region', type=ValidRegionParamType(),
              help="The region where the application is deployed. Default "
                   "value is the one from syndicate config")
@click.option('--attributes', type=(str, str), multiple=True,
              help="SNS application attributes")
@verbose_option
@click.pass_context
@timeit()
def sns_application(ctx, **kwargs):
    """Generates sns application deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = SNSApplicationGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"SNS application '{kwargs['resource_name']}' was added "
                  f"successfully")
    return OK_RETURN_CODE


@meta.command(name="cognito-user-pool")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Cognito user pool name")
# @click.option('--region', type=ValidRegionParamType(), required=True,
#               help="The region where the user pool is created")
@click.option('--auto-verified-attributes',
              cls=MultiWordOption,
              type=click.Choice(['phone_number', 'email']),
              help="The attributes to be auto-verified. "
                   "Default value is email", multiple=True)
@click.option('--sns-caller-arn',
              cls=MultiWordOption, type=str,
              help="The arn of the IAM role in your account which Cognito "
                   "will use to send SMS messages. Required if 'phone_number' "
                   "in 'auto-verified-attributes' is specified")
@click.option('--username-attributes',
              cls=MultiWordOption,
              type=click.Choice(['phone_number', 'email']),
              help="Specifies whether email addresses or phone numbers can "
                   "be specified as usernames when a user signs up. Default "
                   "value is email", multiple=True)
@click.option('--custom-attributes',
              cls=MultiWordOption, type=(str, str), multiple=True,
              help="A list of custom attributes: (name type)")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def cognito_user_pool(ctx, **kwargs):
    """Generates cognito user pool deployment resource template"""
    if 'phone_number' in kwargs['auto_verified_attributes'] \
            and not kwargs.get('sns_caller_arn'):
        raise click.MissingParameter("Sns caller IAM role arn is required when"
                                     " 'phone_number' is specified in "
                                     "'auto-verified-attributes'",
                                     param_type='option',
                                     param_hint='sns-caller-arn')

    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CognitoUserPoolGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Cognito user pool '{kwargs['resource_name']}' was added "
                  f"successfully")
    return OK_RETURN_CODE


@meta.command(name="cognito-federated-pool")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Cognito federated pool name")
@click.option('--auth-role',
              cls=MultiWordOption, type=str,
              help="IAM role for authorized users")
@click.option('--unauth-role',
              cls=MultiWordOption, type=str,
              help="IAM role for unauthorized users")
@click.option('--open-id-providers',
              cls=MultiWordOption, type=str, multiple=True,
              help="A list of OpenID Connect providers")
@click.option('--provider-name',
              cls=MultiWordOption, type=str,
              help="Developer provider name")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def cognito_federated_pool(ctx, **kwargs):
    """Generates cognito federated pool deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CognitoFederatedPoolGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Cognito federated pool '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='batch-compenv')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Batch compute environment name")
@click.option('--compute-environment-type',
              cls=MultiWordOption,
              type=click.Choice(['MANAGED', 'UNMANAGED']),
              help="The type of compute environment. "
                   "Default value is 'MANAGED'")
@click.option('--allocation-strategy',
              cls=MultiWordOption,
              type=click.Choice(['BEST_FIT', 'BEST_FIT_PROGRESSIVE',
                                 'SPOT_CAPACITY_OPTIMIZED']),
              help="The allocation strategy to use for the compute resource "
                   "if not enough instances of the best fitting instance type "
                   "can be allocated")
@click.option('--state', type=click.Choice(['ENABLED', 'DISABLED']),
              help="The state of compute environment")
@click.option('--service-role',
              cls=MultiWordOption, type=str,
              help="The full Amazon Resource Name (ARN) of the IAM role that "
                   "allows Batch to make calls to other Amazon Web Services "
                   "services on your behalf. If not specified, role "
                   "'AWSBatchServiceRole' will be taken if it exists, if "
                   "doesn't it'll be created")
@click.option('--type',
              type=click.Choice(['EC2', 'SPOT', 'FARGATE', 'FARGATE_SPOT']),
              help="The type of compute environment. Default value is EC2")
@click.option('--minv-cpus',
              cls=MultiWordOption, type=click.IntRange(min=0),
              help='The minimum number of Amazon EC2 vCPUs that an '
                   'environment should maintain. Default value is 0')
@click.option('--maxv-cpus',
              cls=MultiWordOption, type=click.IntRange(min=1),
              help="The maximum number of Amazon EC2 vCPUs that a compute "
                   "environment can reach. Default value is 8")
@click.option('--desiredv-cpus',
              cls=MultiWordOption, type=int,
              help="The desired number of Amazon EC2 vCPUS in the compute "
                   "environment. Default value is 1")
@click.option('--instance-types',
              cls=MultiWordOption, type=str, multiple=True,
              help="The instances types that can be launched. Default value "
                   "is 'optimal'")
@click.option('--security-group-ids',
              cls=MultiWordOption, type=str,
              multiple=True, required=True,
              help="The Amazon EC2 security groups associated with instances "
                   "launched in the compute environment")
@click.option('--subnets', type=str, multiple=True, required=True,
              help="The VPC subnets where the compute resources are launched")
@click.option('--instance-role',
              cls=MultiWordOption, type=str,
              help="The Amazon ECS instance profile applied to Amazon EC2 "
                   "instances in a compute environment")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def batch_compenv(ctx, **kwargs):
    """Generates batch compenv deployment resources template"""
    if kwargs.get('type') != 'FARGATE':
        if not kwargs.get('instance_role'):
            raise click.MissingParameter(
                "'instance-role' is required if batch compenv type "
                "ISN'T 'FARGATE'",
                param_type='option',
                param_hint='instance-role')
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = BatchCompenvGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Batch compute environment '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='batch-jobdef')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help='Batch job definition name')
@click.option('--job-definition-type',
              cls=MultiWordOption, required=True,
              type=click.Choice(['container', 'multinode']),
              help='The type of job definition')
@click.option('--image', type=str,
              help='The image used to start a container. '
                   'Default value is \'alpine\'')
@click.option('--job-role-arn',
              cls=MultiWordOption, type=str,
              help='The ARN of the IAM role that the container can assume for '
                   'AWS permissions')
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def batch_jobdef(ctx, **kwargs):
    """Generates batch job definition deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = BatchJobdefGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f'Batch job definition \'{kwargs["resource_name"]}\' was '
                  f'added successfully')
    return OK_RETURN_CODE


@meta.command(name="batch-jobqueue")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Batch job queue name")
@click.option('--state', type=click.Choice(["ENABLED", "DISABLED"]),
              help="The state of the job queue. Default value is 'ENABLED'")
@click.option('--priority', type=int, help="The priority of the job queue. "
                                           "Default value is 1")
@click.option('--compute-environment-order',
              cls=MultiWordOption, type=(int, str), multiple=True,
              help="The set of compute environments mapped to a job queue and "
                   "their order relative to each other. (order, compute_env)")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def batch_jobqueue(ctx, **kwargs):
    """Generates batch job queue deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = BatchJobqueueGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Batch job queue '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="cloudwatch-alarm")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Cloudwatch alarm name")
@click.option('--metric-name',
              cls=MultiWordOption, type=str, required=True,
              help="The metric's name")
@click.option('--namespace', type=str, required=True,
              help="The namespace for the metric associated with the alarm")
@click.option('--description', type=str, help="The description for the alarm")
@click.option('--period', type=click.IntRange(min=1),
              help="The period in seconds over which the specified statistic "
                   "is applied. Valid values are 10, 30 and any multiple"
                   " of 60. Default value is 1200")
@click.option('--evaluation-periods',
              cls=MultiWordOption, type=click.IntRange(min=1),
              help="The number of periods over which data is compared to the "
                   "specified threshold. Default value is 1")
@click.option('--threshold', type=float,
              help="The value to compare with the specified statistic. "
                   "Default value is 1.0")
@click.option('--comparison-operator',
              cls=MultiWordOption,
              type=click.Choice(['GreaterThanOrEqualToThreshold',
                                 'GreaterThanThreshold',
                                 'LessThanThreshold',
                                 'LessThanOrEqualToThreshold',
                                 'LessThanLowerOrGreaterThanUpperThreshold',
                                 'LessThanLowerThreshold',
                                 'GreaterThanUpperThreshold']),
              help="An arithmetic operator to use when comparing the specified"
                   " statistic and threshold. The specified statistic value is"
                   " used as the first operand. Default value is "
                   "'GreaterThanOrEqualToThreshold'")
@click.option('--statistic', type=click.Choice(["SampleCount", "Average",
                                                'Sum', 'Minimum', 'Maximum']),
              help="The statistic for the metric associated with the alarm,"
                   "other than percentile. For percentile statistic use "
                   "'ExtendedStatistic'. Default value is 'SampleCount'")
@click.option('--sns-topics',
              cls=MultiWordOption, type=str, multiple=True,
              help="The sns topics to execute when the alarm goes to an ALARM "
                   "state from any other state")
@click.option('--lambdas', type=str, multiple=True,
              help="The lambdas to execute when the alarm goes to an ALARM "
                   "state from any other state. Use `:` after lambda name to "
                   "specify alias or version")
@click.option('--ssm-response-plan',
              cls=MultiWordOption, type=str, multiple=True,
              help="The response plan name to execute when the alarm goes to "
                   "an ALARM state from any other state")
@click.option('--evaluate-low-sample-count-percentile',
              cls=MultiWordOption,
              type=click.Choice(["evaluate", "ignore"]),
              help="Only for percentiles-based alarms. Use 'ignore' and the "
                   "alarm state remains unchanged during periods with "
                   "insufficient data points for statistical significance. If "
                   "'evaluate' is specified (or parameter is omitted), the "
                   "alarm is always assessed and "
                   "may change state regardless of data point availability")
@click.option('--datapoints', type=click.IntRange(min=1),
              help="The number of datapoints that must be breaching to "
                   "trigger the alarm")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def cloudwatch_alarm(ctx, **kwargs):
    """Generates Cloudwatch alarm deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CloudWatchAlarmGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Cloudwatch alarm '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="cloudwatch-event-rule")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="Cloudwatch event rule name")
@click.option('--rule-type', cls=MultiWordOption,
              required=True, help="Cloudwatch event rule type",
              type=click.Choice(['schedule', 'ec2', 'api_call']))
@click.option('--expression', type=str,
              help="Rule expression (cron schedule). Valuable only if "
                   "rule_type is 'schedule'")
@click.option('--aws-service',
              cls=MultiWordOption, type=str,
              help="The name of AWS service which the rule listens to. "
                   "Required only if rule_type is 'api_call'")
@click.option('--region', type=ValidRegionParamType(allowed_all=True),
              help="The region where the rule is deployed. Default value is "
                   "the one from syndicate config")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def cloudwatch_event_rule(ctx, **kwargs):
    """Generates Cloudwatch event rule deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CloudwatchEventRuleGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"Cloudwatch event rule '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="eventbridge-rule")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="EventBridge rule name")
@click.option('--rule-type', cls=MultiWordOption,
              required=True, help="EventBridge rule type",
              type=click.Choice(['schedule', 'ec2', 'api_call']))
@click.option('--expression', type=str,
              help="Rule expression (cron schedule). Valuable only if "
                   "rule-type is 'schedule'")
@click.option('--aws-service',
              cls=MultiWordOption, type=str,
              help="The name of AWS service which the rule listens to. "
                   "Required only if rule-type is 'api_call'")
@click.option('--region', type=ValidRegionParamType(allowed_all=True),
              help="The region where the rule is deployed. Default value is "
                   "the one from syndicate config")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def eventbridge_rule(ctx, **kwargs):
    """Generates EventBridge rule deployment resources-template
    claiming compatibility with Cloudwatch event rule generator"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = EventBridgeRuleGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"EventBridge rule '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="documentdb-cluster")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="DocumentDB cluster name")
@click.option('--master-username',
              cls=MultiWordOption, type=str, required=True,
              help="DocumentDB login ID for the master user")
@click.option('--master-password',
              cls=MultiWordOption, type=str, required=True,
              help="The password for master user")
@click.option('--port', type=int,
              help="The port number on which the instances in the cluster "
                   "accept connections. Default value is 27017")
@click.option('--vpc-security-group-ids',
              cls=MultiWordOption, type=str, multiple=True,
              help="A list of EC2 VPC security groups to associate with this "
                   "cluster. If not specified, default security group is used")
@click.option('--availability-zones',
              cls=MultiWordOption, type=str, multiple=True,
              help="A list of Amazon EC2 Availability Zones that instances in "
                   "the cluster can be created in. "
                   "If not specified default is used")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def documentdb_cluster(ctx, **kwargs):
    """Generates documentdb cluster deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DocumentDBClusterGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"DocumentDB cluster '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="documentdb-instance")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="DocumentDB instance name")
@click.option('--cluster-identifier',
              cls=MultiWordOption, type=str, required=True,
              help="The identifier of the cluster that the instance will "
                   "belong to")
@click.option('--instance-class',
              cls=MultiWordOption, type=str,
              help="The compute and memory capacity of the instance. Default "
                   "value is 'db.r5.large'")
@click.option('--availability-zone',
              cls=MultiWordOption, type=str,
              help="The Amazon EC2 Availability Zone that the instance is "
                   "created in. If not specified a random zone it the "
                   "endpoint's region is set")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def documentdb_instance(ctx, **kwargs):
    """Generates documentdb instance deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DocumentDBInstanceGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"DocumentDB instance '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name='firehose')
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help='Kinesis Data Firehose delivery stream name')
@click.option('--stream-type',
              cls=MultiWordOption,
              type=click.Choice(['DirectPut', 'KinesisStreamAsSource']),
              default='DirectPut', is_eager=True,
              help='The delivery stream type.')
@click.option('--kinesis-stream-arn',
              type=str, cls=OptionCombined,
              required_if='stream-type',
              required_if_values=['KinesisStreamAsSource'],
              help='The ARN of the source Kinesis data stream. [Required if '
                   'stream-type is \'KinesisStreamAsSource\']')
@click.option('--kinesis-stream-role',
              type=str, cls=OptionCombined,
              required_if='stream-type',
              required_if_values=['KinesisStreamAsSource'],
              help='The role name that provides access to the Kinesis data '
                   'stream source. [Required if stream-type is '
                   '\'KinesisStreamAsSource\']')
@click.option('--destination-role',
              cls=MultiWordOption, type=str, required=True,
              help='The role name that provides access to the Kinesis data '
                   'stream destination S3 bucket.')
@click.option('--destination-bucket',
              cls=MultiWordOption, type=str, required=True,
              help='The Kinesis data stream destination S3 bucket name.')
@click.option('--compression-format',
              cls=MultiWordOption, type=click.Choice(
             ['UNCOMPRESSED', 'GZIP', 'ZIP', 'Snappy', 'HADOOP_SNAPPY']),
              default='UNCOMPRESSED',
              help='The compression format.')
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def firehose(ctx, **kwargs):
    """Generates Kinesis Data Firehose delivery stream deployment resources
    template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = FirehoseGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"The Kinesis Data Firehose delivery stream "
                  f"'{kwargs['resource_name']}' was added successfully")
    return OK_RETURN_CODE


@meta.command(name="eventbridge-schedule")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="EventBridge scheduler name")
@click.option('--schedule-expression',
              cls=MultiWordOption, type=str, required=True,
              help="The expression that defines when the schedule runs. "
                   "The following formats are supported: "
                   "at(yyyy-mm-ddThh:mm:ss); rate(value unit); cron(fields)")
@click.option('--target-arn',
              cls=MultiWordOption, type=str, required=True,
              help="The complete service ARN, including the API operation, in "
                   "the following format: "
                   "`arn:aws:scheduler:::aws-sdk:service:apiAction`. "
                   "For example: arn:aws:scheduler:::aws-sdk:sqs:sendMessage")
@click.option('--role-arn',
              cls=MultiWordOption, type=str, required=True,
              help="The execution role ARN you want to use for the target. "
                   "This role must have the permissions to call the "
                   "API operation you want your schedule to target")
@click.option('--mode', type=click.Choice(['OFF', 'FLEXIBLE']), default='OFF',
              help="Determines whether the schedule is invoked within a "
                   "flexible time window")
@click.option('--maximum-window-in-minutes',
              type=click.IntRange(min=5),
              cls=OptionCombined, required_if='mode',
              required_if_values=['FLEXIBLE'],
              help="The maximum time window during which a schedule can be "
                   "invoked")
@click.option('--description', type=str, help="Schedule description")
@click.option('--schedule-expression-timezone',
              cls=MultiWordOption, type=str,
              help="The timezone in which the scheduling expression "
                   "is evaluated.")
@click.option('--group-name',
              cls=MultiWordOption, type=str,
              help="The name of the schedule group to associate with this "
                   "schedule. By default, the default schedule group is used.")
@click.option('--kms-key-arn',
              cls=MultiWordOption, type=str,
              help="ARN for the customer managed KMS key that scheduler "
                   "will use to encrypt and decrypt data")
@click.option('--state', type=click.Choice(['ENABLED', 'DISABLED']),
              help="Specifies whether the schedule is enabled or disabled")
@click.option('--start-date',
              cls=MultiWordOption, type=str,
              help=" A date in ISO 8601 or UTC, after which the schedule "
                   "can begin invoking its target")
@click.option('--end-date',
              cls=MultiWordOption, type=str,
              help="A date in ISO 8601 or UTC, before which the schedule "
                   "can invoke its target")
@click.option('--dead-letter-arn',
              cls=MultiWordOption, type=str,
              help="SQS queue ARN that will be as the destination "
                   "for the dead-letter queue.")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def eventbridge_schedule(ctx, **kwargs):
    """Generates eventbridge scheduler deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    generator = EventBridgeScheduleGenerator(**filtered_kwargs)
    _generate(generator)
    USER_LOG.info(f"EventBridge scheduler '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="rds-db-cluster")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="DB cluster name")
@click.option('--engine', default='aurora-postgresql',
              type=click.Choice(['aurora-postgresql', 'aurora-mysql']),
              help="Engine type. Default type: aurora-postgresql")
@click.option('--engine-version',
              cls=MultiWordOption, type=str,
              help="Engine version")
@click.option('--master-username',
              cls=MultiWordOption, type=str, required=True,
              help="DB login ID for the master user")
@click.option('--master-password',
              cls=MultiWordOption, type=str,
              callback=partial(
                  validate_incompatible_options,
                  incompatible_options=['manage-master-password']),
              help="The password for master user. Can't be specified if "
                   "manage-master-password is turned on")
@click.option('--database-name',
              cls=MultiWordOption, type=str, required=True,
              help="Database name")
@click.option('--port', type=int,
              help="The port number on which the instances in the cluster "
                   "accept connections. Default value is 3306 for MySQL and "
                   "5432 for PostgreSQL")
@click.option('--manage-master-password',
              cls=MultiWordOption, type=bool, is_eager=True,
              help="Indicates whether to manage the master user password with "
                   "AWS Secrets Manager")
@click.option('--iam-db-auth',
              cls=MultiWordOption, type=bool,
              help="Indicates whether to enable IAM Database Authentication")
@click.option('--vpc-security-group-ids',
              cls=MultiWordOption, type=str, multiple=True,
              help="A list of EC2 VPC security groups to associate with this "
                   "cluster. If not specified, default security group is used")
@click.option('--db-subnet-group',
              cls=MultiWordOption, type=str,
              help="RDS subnet group name to associate with the DB cluster")
@click.option('--availability-zones',
              cls=MultiWordOption, type=str, multiple=True,
              help="A list of Amazon EC2 Availability Zones that instances in "
                   "the cluster can be created in. "
                   "If not specified default is used")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def rds_db_cluster(ctx, **kwargs):
    """Generates RDS DB cluster deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = RDSDBClusterGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"RDS DB cluster '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


@meta.command(name="rds-db-instance")
@return_code_manager
@click.option('--resource-name',
              cls=MultiWordOption, type=str, required=True,
              help="DB instance name")
@click.option('--instance-class',
              cls=MultiWordOption, type=str, required=True,
              help="DB instance class")
@click.option('--cluster-name',
              cls=MultiWordOption, type=str,
              callback=partial(
                  validate_incompatible_options,
                  incompatible_options=
                  RDS_INSTANCE_DB_CLUSTER_INCOMPATIBLE_OPTIONS),
              help="RDS DB cluster name to link the instance with")
@click.option('--engine', type=str, is_eager=True,
              help="Engine type")
@click.option('--engine-version',
              cls=MultiWordOption, type=str, is_eager=True,
              help="Engine version")
@click.option('--master-username',
              cls=MultiWordOption, type=str, is_eager=True,
              help="DB login ID for the master user")
@click.option('--master-password',
              cls=MultiWordOption, type=str, is_eager=True,
              help="The password for master user")
@click.option('--database-name',
              cls=MultiWordOption, type=str, is_eager=True,
              help="Database name")
@click.option('--port', type=int, is_eager=True,
              help="The port number on which the instances in the cluster "
                   "accept connections")
@click.option('--publicly-accessible',
              cls=MultiWordOption, type=bool,
              help="Specifies the accessibility options for the DB instance.")
@click.option('--vpc-security-group-ids',
              cls=MultiWordOption, type=str, multiple=True,
              is_eager=True,
              help="A list of EC2 VPC security groups to associate with this "
                   "cluster. If not specified, default security group is used")
@click.option('--availability-zone',
              cls=MultiWordOption, type=str, is_eager=True,
              help="Amazon EC2 Availability Zone that instances can be "
                   "created in. If not specified default is used")
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@click.pass_context
@timeit()
def rds_db_instance(ctx, **kwargs):
    """Generates RDS DB instance deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = RDSDBInstanceGenerator(**kwargs)
    _generate(generator)
    USER_LOG.info(f"RDS DB cluster '{kwargs['resource_name']}' was "
                  f"added successfully")
    return OK_RETURN_CODE


def _generate(generator: BaseConfigurationGenerator):
    """Just some common actions for this module are gathered in here"""
    try:
        generator.write()
    except AbortedError as e:
        raise click.Abort(e)
    except Exception as e:
        if isinstance(e, SyndicateBaseError):
            raise click.BadParameter(e)
        raise Exception(f"An unexpected error occurred: {e}")




from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import EC2_INSTANCE_TYPE


class EC2InstanceGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = EC2_INSTANCE_TYPE
    CONFIGURATION = {
        'key_name': None,
        'image_id': None,
        'instance_type': 't2.micro',
        'availability_zone': None,
        'subnet_id': None,
        'userdata_file': None,
        'disableApiTermination': True,
        'iam_role': None,
        'security_group_names': list,
        'security_group_ids': list,
        'tags': dict
    }

    def __init__(self, **kwargs):
        if 'disable_api_termination' in kwargs:
            kwargs['disableApiTermination'] = kwargs.pop(
                'disable_api_termination')
        super().__init__(**kwargs)
